"
COPYRIGHT (c) 2020 LabWare
"
"{ Package: 'stx:goodies/monticello/tonel' }"

"{ NameSpace: Smalltalk }"

MCWriter subclass:#TonelWriter
	instanceVariableNames:'snapshot packageDir directoryReference'
	classVariableNames:''
	poolDictionaries:''
	category:'SCM-Monticello-Tonel-Core'
!

TonelWriter comment:'I''m a monticello writer for tonel format, writing
 - a package per directory
 - a class per file
 - a set of extensions to a single class per file (for example, all extensions of a package to String will be in a single file)

I''m created on a file reference to a directory where the package will be written.

[[[
TonelWriter on: (''someDirectory'' asFileReference ensureCreateDirectory)
]]]

My main methods are
- ==#writeVersion:== that receives as argument a monticello version to write, from where I''ll extract the corresponding monticello snapshot.
- ==#writeSnapshot:== that receives as argument a monticello snapshot to write, from where I''ll write all the contained definitions.

I also provide a way to easily export a single class in the Tonel format to a stream. 

[[[
	TonelWriter sourceCodeOf: self.
	
	(FileSystem memory / ''test.st'') writeStreamDo: [ :s | TonelWriter exportClass: self on: s ]; yourself.
]]]

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Implementation details

Notice that while writing, if the written package/snapshot already exists in the directory I''ll overwrite it (i.e., remove it and recreate it).'
!

!TonelWriter class methodsFor:'documentation'!

copyright
"
COPYRIGHT (c) 2020 LabWare

"
! !

!TonelWriter class methodsFor:'instance creation'!

on: directoryReference
    ^ self new
        directoryReference: directoryReference;
        yourself
! !

!TonelWriter class methodsFor:'accessing'!

classLabel
    ^ 'Class'
!

extensionLabel
    ^ 'Extension'
!

readerClass
    ^ TonelReader
!

traitLabel
    ^ 'Trait'
! !

!TonelWriter class methodsFor:'writing'!

fileOut: aVersion on: aStream
    (self on: aStream) writeVersion: aVersion
    
! !

!TonelWriter class methodsFor:'writing - class'!

exportClass: aClass on: aStream
    "I take a class and a stream as parameter and export the class in Tonel format into the stream."
    
    ^ self new exportClass: aClass on: aStream
!

sourceCodeOf: aClass
    "I take as a parameter a class and I return it's tonel export as a String."
    
    ^ String streamContents: [ :aStream | self new exportClass: aClass on: aStream ]
! !

!TonelWriter methodsFor:'accessing'!

directoryReference
    ^ directoryReference
!

directoryReference: aFileReference
    directoryReference := aFileReference
!

fileUtils
    ^ TonelFileUtils current
!

packageDir
    ^ packageDir 
        ifNotNil: [ self fileUtils directoryFromPath: packageDir relativeTo: self directoryReference ]
        ifNil: [ self directoryReference  ]
!

snapshot: anObject
    snapshot := anObject
! !

!TonelWriter methodsFor:'private'!

classNameFor: aMethodDefinition parent: aClassDefinition
    aClassDefinition ifNil: [ ^ aMethodDefinition fullClassName ].
    ^ aMethodDefinition classIsMeta
        ifFalse: [ aMethodDefinition className ]
        ifTrue: [ 
            aClassDefinition isTraitDefinition
                ifFalse: [aMethodDefinition className, ' class']
                ifTrue: [aMethodDefinition className, ' classSide'] ]
!

createDefaultOrganizationFrom: aCollection
    "answers a default organisation for the cases where there are none"

    "simplest case, I answer the clas definition"
    snapshot definitions 
        detect: #isClassDefinition
        ifFound: [ :each | ^ MCOrganizationDefinition categories: { each category } ].
    
    ^ self createDefaultOrganizationFromDefinition: (snapshot definitions
        detect: #isMethodDefinition
        ifNone: [ self error: 'cannot determine package name from empty snapshot' ])
!

createDefaultOrganizationFromDefinition: aMethodDefinition

    ^ MCOrganizationDefinition categories: (self packageNameForMethodDefinition: aMethodDefinition)
!

fileNameFor: aClassDefinition
    ^ String streamContents: [ :aStream | 
        aStream 
            << aClassDefinition className
            << '.' << (self typeOf: aClassDefinition) asLowercase
            << '.st'  ]
!

newLine
    OperatingSystem isMSWINDOWSlike ifTrue: [
        ^ Character return , Character linefeed
    ].
    OperatingSystem isOSXlike ifTrue: [
        ^ Character return asString
    ].
    ^ Character linefeed asString

    "Modified: / 07-07-2020 / 11:49:43 / Jan Vrany <jan.vrany@labware.com>"
!

packageNameForMethodDefinition: aMethodDefinition
    | category |
    
    self assert: aMethodDefinition category first = $*.
    category := aMethodDefinition category allButFirst.
    ^(MCWorkingCopy allManagers
        detect: ((category allSatisfy: #isLowercase)
            ifTrue: [[: wc| category beginsWith: wc packageName asLowercase]]
            ifFalse: [[: wc| category beginsWith: wc packageName]])) packageName
!

selectorIsComplete: keywords in: aString
    keywords inject: 1 into: [ :start :key | | startOfKey | 
        startOfKey := aString 
            findString: key
            startingAt: start 
            caseSensitive: true.
        startOfKey = 0 ifTrue: [ ^ false ].
        startOfKey + key size ].
    ^ true

    "Modified: / 09-07-2020 / 20:35:37 / Jan Vrany <jan.vrany@labware.com>"
!

skipComment: aStream
    "I assume I'm on top of the begining of a comment"
    aStream skip: 1.
    [ aStream atEnd not 
        and: [ aStream next ~= $" or: [ aStream peek = $" ] ] ]
    whileTrue.	
!

skipSeparators: aStream
    [ aStream peek isSeparator ]
    whileTrue: [ aStream skip: 1 ]. 
!

splitMethodSource: aMethodDefinition into: aBlock
    | keywords source declaration |
    
    keywords := aMethodDefinition selector keywords.
    source := aMethodDefinition source readStream.
    "Skip spaces"
    (source peek isSeparator) ifTrue: [ self skipSeparators: source ].
    "Skip comments"
    (source peek = $") ifTrue: [ self skipComment: source ]. 
    "Parse declaration"
    declaration := String new writeStream.
    [ (self selectorIsComplete: keywords in: declaration contents) not 
        or: [ ':+-/\*~<>=@,%|&?!!' includes: declaration contents trimRight last ] ]
    whileTrue: [ 
        "stop infinite loop if no match was found"
        source atEnd ifTrue: [ TonelWriteError signal: 'Cannot find selector in source for ', aMethodDefinition asString ].
        "get separators"
        [ source atEnd not and: [ source peek isSeparator ] ]
            whileTrue: [ declaration nextPut: source next ].
        "take next word until we find a separator or a dot"
        [ source atEnd not and: [ source peek ~= $. and: [ source peek isSeparator not ] ] ]
            whileTrue: [ declaration nextPut: source next ] ].
    aBlock 
        value: (declaration contents trimLeft withLineEndings: self newLine)
        value: (source upToEnd withLineEndings: self newLine)

    "Modified: / 25-05-2020 / 14:54:13 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

toSTON: anObject
    ^ String streamContents: [ :outStream | 
        (TonelSTONWriter on: outStream) 
            newLine: self newLine;
            nextPut: anObject ]
!

typeOf: aClassDefinition
    (self isClass: aClassDefinition) ifTrue: [ ^ self class classLabel ].
    (self isTrait: aClassDefinition) ifTrue: [ ^ self class traitLabel ].

    TonelShouldIgnore signal
! !

!TonelWriter methodsFor:'private definitions'!

commentOf: aClassDefinition
    ^ (aClassDefinition comment 
        copyReplaceAll: '"' 
        with: '""')
        withLineEndings: self newLine
!

methodDefinitionOf: aMethodDefinition
    ^ self toSTON: (OrderedDictionary new 
        at: #category put: aMethodDefinition category; 
        yourself)	
    
!

typeClassDefinitionOf: aClassDefinition
    | definition |
    
    definition := OrderedDictionary new 
        at: #name put: aClassDefinition className; 
        at: #superclass put: aClassDefinition superclassName;
        yourself.

    aClassDefinition type = #normal ifFalse: [ 
        definition at: #type put: aClassDefinition type ].
    
    aClassDefinition hasTraitComposition ifTrue: [ 
        definition at: #traits put: aClassDefinition traitCompositionString ].
    
    aClassDefinition hasClassTraitComposition ifTrue: [ 
        definition at: #classTraits put: aClassDefinition classTraitCompositionString ].
    
    (aClassDefinition variables select: #isInstanceVariable thenCollect: #name)
        ifNotEmpty: [ :vars | definition at: #instVars put: vars asArray ].

    (aClassDefinition variables select: #isClassVariable thenCollect: #name)
        ifNotEmpty: [ :vars | definition at: #classVars put: vars asArray ].
        
    (aClassDefinition variables select: #isPoolImport thenCollect: #name)
        ifNotEmpty: [ :vars | definition at: #pools put: vars asArray ].
        
    (aClassDefinition variables select: #isClassInstanceVariable thenCollect: #name)
        ifNotEmpty: [ :vars | definition at: #classInstVars put: vars asArray ].

    definition 		
        at: #category put: aClassDefinition category.
    
    ^ self toSTON: definition
!

typeDefinitionOf: aClassDefinition
    (self isTrait: aClassDefinition) 
        ifTrue: [ ^ self typeTraitDefinitionOf: aClassDefinition ].
    ^ self typeClassDefinitionOf: aClassDefinition
!

typeTraitDefinitionOf: aClassDefinition
    | definition |
    
    definition := OrderedDictionary new 
        at: #name put: aClassDefinition className; 
        yourself.

    aClassDefinition variables isEmptyOrNil ifFalse: [ 
        (aClassDefinition variables select: #isInstanceVariable thenCollect: #name)
            ifNotEmpty: [ :vars | definition at: #instVars put: vars asArray ].
        (aClassDefinition variables select: #isClassInstanceVariable thenCollect: #name)
            ifNotEmpty: [ :vars | definition at: #classInstVars put: vars asArray ] ].

    aClassDefinition hasTraitComposition ifTrue: [ 
        definition		
            at: #traits 
            put: aClassDefinition traitCompositionString ].
        
    aClassDefinition hasClassTraitComposition ifTrue: [ 
        definition		
            at: #classTraits 
            put: aClassDefinition classTraitCompositionString ].
    
    definition at: #category put: aClassDefinition category.
        
    ^ self toSTON: definition
! !

!TonelWriter methodsFor:'private testing'!

isClass: aClassDefinition
    ^ aClassDefinition class = MCClassDefinition
!

isTrait: aClassDefinition
    ^ aClassDefinition class = MCTraitDefinition 
! !

!TonelWriter methodsFor:'private writing'!

writeClassDefinition: aClassDefinition on: aStream
    | nl |
    nl := self newLine.
    
    aClassDefinition hasComment 
        ifTrue: [ 
            aStream 
                << '"' << nl
                << (self commentOf: aClassDefinition) << nl
                << '"' << nl ].
    aStream
        << (self typeOf: aClassDefinition) 
        << ' ' << (self typeDefinitionOf: aClassDefinition ) << nl
!

writeClassSideMethodDefinitions: aClassDefinition on: aStream
    ((snapshot definitions 
        select: [ :each | 
            each isMethodDefinition 
            and: [ each className = aClassDefinition className
            and: [ each classIsMeta ] ] ])
        sorted: [ :a :b | a selector < b selector ])
        do: [ :each | 
            self writeMethodDefinition: each parent: aClassDefinition on: aStream ]
!

writeInstanceSideMethodDefinitions: aClassDefinition on: aStream
    ((snapshot definitions 
        select: [ :each | 
            each isMethodDefinition 
            and: [ each className = aClassDefinition className
            and: [ each classIsMeta not ] ] ])
        sorted: [ :a :b | a selector < b selector ])
        do: [ :each | 
            self writeMethodDefinition: each parent: aClassDefinition on: aStream ]

    
    
!

writeMethodDefinition: aMethodDefinition on: aStream
    ^ self 
        writeMethodDefinition: aMethodDefinition 
        parent: nil 
        on: aStream
!

writeMethodDefinition: aMethodDefinition parent: aClassDefinition on: aStream
    | nl |
    
    nl := self newLine.
    self 
        splitMethodSource: aMethodDefinition 
        into: [ :methodDeclaration :methodBody | | fullClassName |
            fullClassName := self classNameFor: aMethodDefinition parent: aClassDefinition.
            aStream 
                << nl 
                << (self methodDefinitionOf: aMethodDefinition) << nl 
                << fullClassName << ' >> ' << methodDeclaration 
                << ' [' << methodBody << nl << ']' << nl ]
! !

!TonelWriter methodsFor:'writing'!

exportClass: aClass on: aStream
    snapshot := MCSnapshot
        fromDefinitions:
            {aClass asClassDefinition} , (aClass methodDictionary values collect: [ :each | each asMethodDefinition ])
                , (aClass classSide methodDictionary values collect: [ :each | each asMethodDefinition ]).
    self writeClass: aClass asClassDefinition on: aStream.
    ^ aStream

    "Modified: / 06-07-2020 / 21:39:20 / Jan Vrany <jan.vrany@labware.com>"
!

writeClass: aClassDefinition
    [ self fileUtils writeStreamFor: (self fileNameFor: aClassDefinition) in: self packageDir do: [ :aStream | self writeClass: aClassDefinition on: aStream ] ]
        on: TonelShouldIgnore
        do: [ :e | self logCr: 'ignoring: ' , aClassDefinition asString ]
!

writeClass: aClassDefinition on: aStream
    [ self writeClassDefinition: aClassDefinition on: aStream.
    self writeClassSideMethodDefinitions: aClassDefinition on: aStream.
    self writeInstanceSideMethodDefinitions: aClassDefinition on: aStream ]
        on: TonelShouldIgnore
        do: [ :e | Logger info: 'Tonel: ignoring: %1' with: aClassDefinition asString ]

    "Modified: / 07-07-2020 / 11:48:49 / Jan Vrany <jan.vrany@labware.com>"
!

writeExtensionMethods: methods className: className
    | nl |
    nl := self newLine.
    self fileUtils 
        writeStreamFor: (className, '.extension.st') 
        in: self packageDir 
        do: [ :s |
            s << 'Extension ' << (self toSTON: (Dictionary with: #name -> className) ) << nl.
            (methods sorted: [ :a :b | a selector < b selector ])
                do: [ :each | self writeMethodDefinition: each on: s ] ]

    "Modified: / 05-08-2020 / 16:25:17 / Jan Vrany <jan.vrany@labware.com>"
!

writeMethodExtensions
    | classesAndMethods |
    classesAndMethods := OrderedDictionary new.
    snapshot definitions 
        select: [ :each | each isMethodDefinition and: [ each isExtensionMethod ] ]
        thenDo: [ :each | 
            (classesAndMethods 
                at: each className
                ifAbsentPut: [ OrderedSet new ])
                add: each ].
    classesAndMethods keysAndValuesDo: [ :className :methods | 
        self 
            writeExtensionMethods: methods 
            className: className ]

    "Modified: / 07-07-2020 / 22:34:34 / Jan Vrany <jan.vrany@labware.com>"
!

writePackage: aDefinition

    "/self flag: #todo. "this is lame... removing and rewriting full package needs to be rethink :)"
    (self fileUtils directoryExists: self packageDir) ifTrue: [ 
        self fileUtils deleteAll: self packageDir ].  
    self fileUtils ensureDirectoryExists: self packageDir.
    self fileUtils 
        writeStreamFor: 'package.st' 
        in: self packageDir 
        do: [ :s | 
            s 
                << 'Package ' 
                << (self toSTON: (Dictionary with: #name ->  packageDir asSymbol) ) 
                << self newLine ]

    "Modified: / 31-08-2020 / 16:26:45 / Jan Vrany <jan.vrany@labware.com>"
!

writeSnapshot: aSnapshot
    snapshot := aSnapshot.
    
    "ensure package dirs exists. 
     It has to be just one but well..."
    self writePackage: (snapshot definitions 
        detect: #isOrganizationDefinition
        ifNone: [ self createDefaultOrganizationFrom: snapshot definitions ]).
    "now export classes"
    (snapshot definitions 
        select: #isClassDefinition)
        do: [ :each | self writeClass: each ].
    "... and method extensions"
    self writeMethodExtensions
!

writeVersion: aVersion
    packageDir := aVersion info name.
    self writeSnapshot: aVersion snapshot

    "Modified: / 12-08-2020 / 12:44:17 / Jan Vrany <jan.vrany@labware.com>"
! !

!TonelWriter class methodsFor:'documentation'!

version_HG

    ^ '$Changeset: <not expanded> $'
! !

